APP_WIDTH = 240
APP_HEIGHT = 180

EDITOR_IMAGE_X = 0
EDITOR_IMAGE_Y = 32
EDITOR_IMAGE_NAME = "editor_220x160.png"

TILEMAP_IMAGE_X = 0
TILEMAP_IMAGE_Y = 192
TILEMAP_IMAGE_WIDTH = 64
TILEMAP_IMAGE_HEIGHT = 63

TOOL_SELECT = 0
TOOL_PENCIL = 1
TOOL_RECTB = 2
TOOL_RECT = 3
TOOL_CIRCB = 4
TOOL_CIRC = 5
TOOL_BUCKET = 6

MAX_SOUND_LENGTH = 48
MAX_MUSIC_LENGTH = 32
